import asyncio
from enum import Enum
from glootil import (
    Toolbox,
    DynEnum,
    DynTagValue,
    FixedTagValue,
    FnArg,
    NoneType,
    type_to_schema_type,
    basic_match_raw_tag_value,
    cast_json_type_to_fn_arg_type_or_default as cast_json,
)
from datetime import date

import pytest

all_months = [
    ("01", "January"),
    ("02", "February"),
    ("03", "March"),
    ("04", "April"),
    ("05", "May"),
    ("06", "June"),
    ("07", "July"),
    ("08", "August"),
    ("09", "September"),
    ("10", "October"),
    ("11", "November"),
    ("12", "December"),
]


def test_toolbox_constructor():
    t = Toolbox("math", "Math tools", "do operations on numbers")
    assert t.id == "math"
    assert t.name == "Math tools"
    assert t.docs == "do operations on numbers"


def test_tool_no_call_no_args():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.tool
    def f():
        return None

    t = tb.tools[0]

    assert t.name == "f"
    assert t.docs is None
    assert len(t.args) == 0


def test_tool_no_call_no_annotations():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.tool
    def add(a, b):
        return a + b

    t = tb.tools[0]

    assert t.name == "add"
    assert t.docs is None
    assert len(t.args) == 2

    a, b = t.args

    assert a.name == "a"
    assert a.index == 0
    assert a.type is None
    assert a.default_value is None

    assert b.name == "b"
    assert b.index == 1
    assert b.type is None
    assert b.default_value is None

    assert t.return_type is None


def test_tool_no_call():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.tool
    def add(a: int, b: int = 0) -> int:
        "add two numbers together"
        return a + b

    t = tb.tools[0]

    assert t.name == "add"
    assert t.docs == "add two numbers together"
    assert len(t.args) == 2

    a, b = t.args

    assert a.name == "a"
    assert a.index == 0
    assert a.type is int
    assert a.default_value is None

    assert b.name == "b"
    assert b.index == 1
    assert b.type is int
    assert b.default_value == 0

    assert t.return_type is int


def test_tool_call_override_name_and_args():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.tool(
        name="substract", docs="my docs", args={"a": {"name": "Left"}, "b": "Right"}
    )
    def sub(a, b):
        return a - b

    t = tb.tools[0]

    assert t.name == "substract"
    assert t.docs == "my docs"
    assert len(t.args) == 2

    a, b = t.args

    assert a.label == "Left"
    assert b.label == "Right"


def test_enum_class():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum
    class Operation(Enum):
        "the operation to apply"

        ADD = "add"
        SUB = "sub"
        MUL = "mul"
        DIV = "div"

    assert len(tb.enums) == 1

    e = tb.enums[0]

    assert isinstance(e, FixedTagValue)
    assert e.name == "Operation"
    assert e.docs == "the operation to apply"
    assert len(e.variants) == 4

    add, sub, mul, div = e.variants

    assert add.name == "ADD"
    assert add.value == "add"

    assert sub.name == "SUB"
    assert sub.value == "sub"

    assert mul.name == "MUL"
    assert mul.value == "mul"

    assert div.name == "DIV"
    assert div.value == "div"

    assert e.to_info() == {
        "id": e.id,
        "name": e.name,
        "description": e.docs,
        "icon": e.icon,
        "entries": [(v.name, v.value) for v in e.variants],
        "matchHandlerId": f"enum::{e.id}::match",
    }


def test_empty_build_tool_info():
    tb = Toolbox("mytools", "My Tools", "some tools")

    assert tb.build_tool_info() == {
        "ns": "mytools",
        "title": "My Tools",
        "description": "some tools",
        "tools": {},
        "tagValues": {},
    }


def test_type_to_schema_type_int():
    assert type_to_schema_type(int, None) == ("integer", 0)
    assert type_to_schema_type(int, 42) == ("integer", 42)


def test_type_to_schema_type_float():
    assert type_to_schema_type(float, None) == ("number", 0.0)
    assert type_to_schema_type(float, 3.14) == ("number", 3.14)


def test_type_to_schema_type_str():
    assert type_to_schema_type(str, None) == ("string", "")
    assert type_to_schema_type(str, "hello") == ("string", "hello")


def test_type_to_schema_type_bool():
    assert type_to_schema_type(bool, None) == ("boolean", False)
    assert type_to_schema_type(bool, True) == ("boolean", True)


def test_type_to_schema_type_fallthrough():
    assert type_to_schema_type(list, None) == ("string", "")


def test_tool_all_types():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.tool
    def all_types(
        a: int,
        b: float,
        c: bool,
        d: str,
        e: int = 10,
        f: float = 1.5,
        g: bool = True,
        h: str = "hi",
    ):
        pass

    t = tb.tools[0]

    assert t.to_info() == {
        "title": t.name,
        "description": t.docs,
        "schema": {
            "fields": {
                "a": {"type": "integer", "default": 0, "description": None},
                "b": {"type": "number", "default": 0.0, "description": None},
                "c": {"type": "boolean", "default": False, "description": None},
                "d": {"type": "string", "default": "", "description": None},
                "e": {"type": "integer", "default": 10, "description": None},
                "f": {"type": "number", "default": 1.5, "description": None},
                "g": {"type": "boolean", "default": True, "description": None},
                "h": {"type": "string", "default": "hi", "description": None},
            }
        },
        "ui": {
            "prefix": t.name,
            "args": {
                "a": {"prefix": "a"},
                "b": {"prefix": "b"},
                "c": {"prefix": "c"},
                "d": {"prefix": "d"},
                "e": {"prefix": "e"},
                "f": {"prefix": "f"},
                "g": {"prefix": "g"},
                "h": {"prefix": "h"},
            },
        },
        "examples": [],
    }


def test_toolbox_info():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.tool(examples=["Show all types"])
    def all_types(
        a: int,
        b: float,
        c: bool,
        d: str,
        e: int = 10,
        f: float = 1.5,
        g: bool = True,
        h: str = "hi",
    ):
        pass

    t = tb.tools[0]

    @tb.enum(id="CountryEnum", icon="flag")
    class Country(DynEnum):
        "A Country"

        @staticmethod
        def load():
            return [("ARG", "Argentina"), ("URU", "Uruguay")]

    e = tb.enums[0]

    assert tb.build_tool_info() == {
        "ns": "mytools",
        "title": "My Tools",
        "description": "some tools",
        "tools": {
            "all_types": {
                "title": t.name,
                "description": t.docs,
                "schema": {
                    "fields": {
                        "a": {"type": "integer", "default": 0, "description": None},
                        "b": {"type": "number", "default": 0.0, "description": None},
                        "c": {"type": "boolean", "default": False, "description": None},
                        "d": {"type": "string", "default": "", "description": None},
                        "e": {"type": "integer", "default": 10, "description": None},
                        "f": {"type": "number", "default": 1.5, "description": None},
                        "g": {"type": "boolean", "default": True, "description": None},
                        "h": {"type": "string", "default": "hi", "description": None},
                    }
                },
                "ui": {
                    "prefix": t.name,
                    "args": {
                        "a": {"prefix": "a"},
                        "b": {"prefix": "b"},
                        "c": {"prefix": "c"},
                        "d": {"prefix": "d"},
                        "e": {"prefix": "e"},
                        "f": {"prefix": "f"},
                        "g": {"prefix": "g"},
                        "h": {"prefix": "h"},
                    },
                },
                "examples": ["Show all types"],
            }
        },
        "tagValues": {
            "CountryEnum": {
                "id": e.id,
                "name": e.name,
                "description": e.docs,
                "icon": e.icon,
                "matchHandlerId": "enum::CountryEnum::match",
                "loadEntriesHandlerId": "enum::CountryEnum::load_entries",
            }
        },
    }


def test_basic_matcher_empty_pos():
    assert basic_match_raw_tag_value("", []) == (-1, None)


def test_basic_matcher_non_empty_pos_list():
    assert basic_match_raw_tag_value("a", [("b", "c")]) == (0, ("b", "c"))


def gen_one_posibility():
    yield ("b", "c")


def test_basic_matcher_non_empty_pos_gen():
    assert basic_match_raw_tag_value("a", gen_one_posibility()) == (0, ("b", "c"))


def gen_two_posibilities():
    yield ("b", "c")
    yield ("a", "a")


def test_basic_matcher_non_empty_pos_gen_2():
    assert basic_match_raw_tag_value("a", gen_two_posibilities()) == (1, ("a", "a"))


@pytest.mark.asyncio
async def test_enum_class_default_matcher():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum
    class Operation(Enum):
        "the operation to apply"

        ADD = "add"
        SUB = "sub"
        MUL = "mul"
        DIV = "div"

    e = tb.enums[0]

    assert await e.closest_match("ad") == ("ADD", "add")
    assert await e.closest_match("mu") == ("MUL", "mul")


@pytest.mark.asyncio
async def test_enum_class_custom_matcher():
    def matcher_always_first(_word, possibilities):
        return 0, next(iter(possibilities))

    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum(matcher=matcher_always_first)
    class Operation(Enum):
        "the operation to apply"

        ADD = "add"
        SUB = "sub"
        MUL = "mul"
        DIV = "div"

    e = tb.enums[0]

    assert e.closest_matcher == matcher_always_first
    assert await e.closest_match("") == ("ADD", "add")
    assert await e.closest_match("mul") == ("ADD", "add")


@pytest.mark.asyncio
async def test_tb_setup_with_sync_state_setup():
    class State:
        def __init__(self):
            self.v = None

        def setup(self):
            self.v = 42

    tb = Toolbox("mytools", "My Tools", "some tools", state=State())
    await tb.setup()
    assert tb.state.v == 42


@pytest.mark.asyncio
async def test_tb_setup_with_async_state_setup():
    class State:
        def __init__(self):
            self.v = None

        async def setup(self):
            await asyncio.sleep(0.010)
            self.v = 42

    tb = Toolbox("mytools", "My Tools", "some tools", state=State())
    await tb.setup()
    assert tb.state.v == 42


@pytest.mark.asyncio
async def test_tb_setup_with_no_state_setup():
    class State:
        def __init__(self):
            self.v = None

    tb = Toolbox("mytools", "My Tools", "some tools", state=State())
    await tb.setup()
    assert tb.state.v is None


@pytest.mark.asyncio
async def test_enum_class_match_handler():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum
    class Operation(Enum):
        "the operation to apply"

        ADD = "add"
        SUB = "sub"
        MUL = "mul"
        DIV = "div"

    e = tb.enums[0]

    assert tb.handlers[e.match_handler_id] == e.match_handler
    match_result = await e.match_handler(dict(value="ad"))
    assert match_result == ("ADD", "add")


def test_provide_arg_int():
    tb = Toolbox("mytools", "My Tools", "some tools")
    arg = FnArg("a", 0, int, 0)
    assert tb.provide_arg(arg, {}) == 0
    assert tb.provide_arg(arg, {"a": 5}) == 5
    assert tb.provide_arg(arg, {"a": 5.5}) == 5


def test_provide_arg_float():
    tb = Toolbox("mytools", "My Tools", "some tools")
    arg = FnArg("a", 0, float, 0)
    assert tb.provide_arg(arg, {}) == 0
    assert tb.provide_arg(arg, {"a": 5}) == 5
    assert tb.provide_arg(arg, {"a": 5.5}) == 5.5
    assert tb.provide_arg(arg, {"a": False}) == 0


def test_provide_arg_bool():
    tb = Toolbox("mytools", "My Tools", "some tools")
    arg = FnArg("a", 0, bool, False)
    assert tb.provide_arg(arg, {}) is False
    assert tb.provide_arg(arg, {"a": True}) is True
    assert tb.provide_arg(arg, {"a": 1}) is True
    assert tb.provide_arg(arg, {"a": 0}) is False


def test_provide_arg_str():
    tb = Toolbox("mytools", "My Tools", "some tools")
    arg = FnArg("a", 0, str, "hi")
    assert tb.provide_arg(arg, {}) == "hi"
    assert tb.provide_arg(arg, {"a": "hello"}) == "hello"
    assert tb.provide_arg(arg, {"a": ""}) == ""
    assert tb.provide_arg(arg, {"a": 0}) == "0"


@pytest.mark.asyncio
async def test_provide_arg_enum():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum
    class Color(Enum):
        RED = "Red"
        GREEN = "Green"
        BLUE = "Blue"

    arg = FnArg("a", 0, Color, Color.RED)
    assert await tb.provide_arg(arg, {}) == Color.RED
    assert await tb.provide_arg(arg, {"a": "BLUE"}) == Color.BLUE
    assert await tb.provide_arg(arg, {"a": "Blue"}) == Color.BLUE
    assert await tb.provide_arg(arg, {"a": "Bl"}) == Color.BLUE
    assert await tb.provide_arg(arg, {"a": "G"}) == Color.GREEN


@pytest.mark.asyncio
async def test_provide_arg_dyn_enum():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum
    class Color(DynEnum):
        @staticmethod
        def load():
            return [
                ("RED", "Red"),
                ("GREEN", "Green"),
                ("BLUE", "Blue"),
            ]

    RED = Color("RED", "Red")
    GREEN = Color("GREEN", "Green")
    BLUE = Color("BLUE", "Blue")

    arg = FnArg("a", 0, Color, RED)
    assert await tb.provide_arg(arg, {}) == RED
    assert await tb.provide_arg(arg, {"a": "BLUE"}) == BLUE
    assert await tb.provide_arg(arg, {"a": "Blue"}) == BLUE
    assert await tb.provide_arg(arg, {"a": "Bl"}) == BLUE
    assert await tb.provide_arg(arg, {"a": "G"}) == GREEN


def test_provide_arg_optional_str():
    tb = Toolbox("mytools", "My Tools", "some tools")
    arg = FnArg("a", 0, str | None, None)
    assert tb.provide_arg(arg, {}) is None
    assert tb.provide_arg(arg, {"a": "hello"}) == "hello"
    assert tb.provide_arg(arg, {"a": ""}) == ""
    assert tb.provide_arg(arg, {"a": 0}) == "0"


def test_provide_arg_optional_str_in_tool():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.tool
    def f(a: str | None = "hi"):
        return a

    @tb.tool
    def g(a: None | str):
        return a

    t = tb.tools[0]
    arg = t.args[0]
    assert arg.type is str
    assert arg._isinstance_type == (str, type(None))

    t = tb.tools[1]
    arg = t.args[0]
    assert arg.type is str
    assert arg._isinstance_type == (type(None), str)


def test_dyn_enum_decoration_no_args():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum
    class Country(DynEnum):
        "A Country"

        @staticmethod
        def load():
            return [("ARG", "Argentina"), ("URU", "Uruguay")]

    assert len(tb.enums) == 1

    e = tb.enums[0]

    assert isinstance(e, DynTagValue)
    assert e.id == "Country"
    assert e.name == "Country"
    assert e.docs == "A Country"
    assert e.icon is None

    assert e.to_info() == {
        "id": e.id,
        "name": e.name,
        "description": e.docs,
        "icon": e.icon,
        "matchHandlerId": "enum::Country::match",
        "loadEntriesHandlerId": "enum::Country::load_entries",
    }


def test_dyn_enum_decoration_args():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum(id="CountryEnum", icon="flag")
    class Country(DynEnum):
        "A Country"

        @staticmethod
        def load():
            return [("ARG", "Argentina"), ("URU", "Uruguay")]

    assert len(tb.enums) == 1

    e = tb.enums[0]

    assert isinstance(e, DynTagValue)
    assert e.id == "CountryEnum"
    assert e.name == "Country"
    assert e.docs == "A Country"
    assert e.icon == "flag"

    assert e.to_info() == {
        "id": e.id,
        "name": e.name,
        "description": e.docs,
        "icon": e.icon,
        "matchHandlerId": "enum::CountryEnum::match",
        "loadEntriesHandlerId": "enum::CountryEnum::load_entries",
    }


@pytest.mark.asyncio
async def test_fn_enum_default_matcher():
    def matcher_always_first(_word, possibilities):
        return 0, next(iter(possibilities))

    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum
    class Month(DynEnum):
        @staticmethod
        def load():
            return all_months

    e = tb.enums[0]

    assert await e.closest_match("1") == ("01", "January")
    assert await e.closest_match("2") == ("02", "February")
    assert await e.closest_match("may") == ("05", "May")


@pytest.mark.asyncio
async def test_fn_enum_custom_matcher():
    def matcher_always_first(_word, possibilities):
        return 0, next(iter(possibilities))

    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum(matcher=matcher_always_first)
    class Month(DynEnum):
        @staticmethod
        def load():
            return all_months

    e = tb.enums[0]

    assert e.closest_matcher == matcher_always_first
    assert await e.closest_match("") == ("01", "January")
    assert await e.closest_match("02") == ("01", "January")
    assert await e.closest_match("May") == ("01", "January")


@pytest.mark.asyncio
async def test_dyn_enum_arg_provider_provides_state():
    class State:
        def get_variants(self):
            return all_months

    tb = Toolbox("mytools", "My Tools", "some tools", state=State())

    @tb.enum
    class Month(DynEnum):
        @staticmethod
        def load(name="Months", state: State | None = None):
            return state.get_variants() if state else []

    e = tb.enums[0]
    en = await e.get_variants()

    assert len(en) == len(all_months)
    assert en == all_months


@pytest.mark.asyncio
async def test_async_dyn_enum():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum
    class Month(DynEnum):
        "my enum description"

        @staticmethod
        async def load():
            await asyncio.sleep(0.010)
            return all_months

    assert len(tb.enums) == 1

    e = tb.enums[0]

    assert isinstance(e, DynTagValue)
    assert e.name == "Month"
    assert e.docs == "my enum description"
    assert e.icon is None
    en = await e.get_variants()
    assert len(en) == len(all_months)
    assert en == all_months


@pytest.mark.asyncio
async def test_dyn_enum_match_handler():
    tb = Toolbox("mytools", "My Tools", "some tools")

    @tb.enum
    class Operation(DynEnum):
        @staticmethod
        def load():
            return [
                ("ADD", "add"),
                ("SUB", "sub"),
                ("MUL", "mul"),
                ("DIV", "div"),
            ]

    e = tb.enums[0]

    assert tb.handlers[e.match_handler_id] == e.match_handler
    match_result = await e.match_handler(dict(value="ad"))
    assert match_result == ("ADD", "add")


def test_cast_json_from_int():
    assert cast_json(10, int, 0) == 10
    assert cast_json(10, float, 0.0) == 10.0
    assert cast_json(10, bool, False) is True
    assert cast_json(0, bool, True) is False
    assert cast_json(5, str, "") == "5"
    assert cast_json(5, date, None) is None


def test_cast_json_from_float():
    assert cast_json(10.5, float, 0) == 10.5
    assert cast_json(10.5, int, 0) == 10
    assert cast_json(10.5, bool, False) is True
    assert cast_json(0.0, bool, True) is False
    assert cast_json(5.5, str, "") == "5.5"
    assert cast_json(5, date, None) is None


def test_cast_json_from_bool():
    assert cast_json(True, bool, None) is True
    assert cast_json(False, bool, None) is False
    assert cast_json(True, int, 0) == 1
    assert cast_json(False, int, 1) == 0
    assert cast_json(True, float, 0) == 1.0
    assert cast_json(False, float, 1) == 0.0
    assert cast_json(True, str, "") == "True"
    assert cast_json(False, str, "") == "False"
    assert cast_json(True, date, None) is None


def test_cast_json_from_none():
    assert cast_json(None, NoneType, 10) is None
    assert cast_json(None, int, 10) == 10
    assert cast_json(None, float, 5.0) == 5.0
    assert cast_json(None, bool, True) is True
    assert cast_json(None, str, "nope") == "nope"
    assert cast_json(None, date, None) is None


def test_cast_json_from_str():
    assert cast_json("10.5", float, 0.0) == 10.5
    assert cast_json("10", int, 0) == 10
    assert cast_json("", bool, True) is False
    assert cast_json("whatev", bool, False) is True
    assert cast_json("hi", str, "") == "hi"
    assert cast_json("2025-03-25", date, None) == date(2025, 3, 25)



import asyncio
import inspect
import typing
from difflib import SequenceMatcher
from enum import Enum
from datetime import date

UnionType = type(str | None)
NoneType = type(None)

# first level is the type of input, second is expected output type
# JSON_CASTER[from][to]
JSON_CASTERS = {
    float: {
        NoneType: lambda _v, _d: None,
        float: lambda v, _d: v,
        str: lambda v, _d: str(v),
        int: lambda v, _d: int(v),
        bool: lambda v, _d: bool(v),
        date: lambda _v, d: cast_date_default(d),
    },
    int: {
        NoneType: lambda _v, _d: None,
        int: lambda v, _d: v,
        str: lambda v, _d: str(v),
        float: lambda v, _d: float(v),
        bool: lambda v, _d: bool(v),
        date: lambda _v, d: cast_date_default(d),
    },
    str: {
        NoneType: lambda _v, _d: None,
        str: lambda v, _d: v,
        int: lambda v, d: try_cast_or(v, int, d),
        float: lambda v, d: try_cast_or(v, float, d),
        bool: lambda v, d: try_cast_or(v, bool, d),
        date: lambda v, d: try_cast_or(v, date.fromisoformat, d, cast_date_default),
    },
    bool: {
        NoneType: lambda _v, _d: None,
        bool: lambda v, _d: v,
        str: lambda v, _d: str(v),
        int: lambda v, _d: int(v),
        float: lambda v, _d: float(v),
        date: lambda _v, d: cast_date_default(d),
    },
    NoneType: {
        NoneType: lambda _v, _d: None,
        bool: lambda _v, d: d,
        str: lambda _v, d: d,
        int: lambda _v, d: d,
        float: lambda _v, d: d,
        date: lambda _v, d: cast_date_default(d),
    },
}


def cast_date_default(v):
    # TODO: handle a way to have dynamic default like "today"
    return v


def identity(v):
    return v


def try_cast_or(v, cast_fn, default, default_caster=identity):
    try:
        return cast_fn(v)
    except Exception as err:
        print("couldn't cast to", cast_fn, "returning default", err)
        return default_caster(default)


def cast_json_type_to_fn_arg_type_or_default(v, target_type, default):
    v_type = type(v)
    caster = JSON_CASTERS.get(v_type, {}).get(target_type)
    if caster:
        return caster(v, default)
    else:
        print(
            "WARN: no caster found from", v_type, "to", target_type, "returning default"
        )
        return default


class FnArg:
    def __init__(self, name, index, type_, default_value):
        self.name = name
        self.index = index
        if type_ and typing.get_origin(type_) is UnionType:
            targs = typing.get_args(type_)
            ta, tb = targs
            if len(targs) == 2 and (ta is NoneType or tb is NoneType):
                self.type = ta if tb is NoneType else tb
                self._isinstance_type = targs
            else:
                raise ValueError("FnArg type Union can only be with None")
        else:
            # assert that it's a type
            assert type_ is None or isinstance(type_, type)
            self.type = type_

            if type_ is None:
                self._isinstance_type = object
            elif type_ is float:
                self._isinstance_type = (int, float)
            else:
                self._isinstance_type = type_

        self.default_value = default_value

    def cast_json_value_or_default(self, v):
        "cast a value that came from json, handles only int, float, bool, null, string"
        return cast_json_type_to_fn_arg_type_or_default(
            v, self.type, self.default_value
        )

    def __str__(self):
        return f"{self.name}[{self.index}]: {self.type} = {self.default_value}"


class ToolArg(FnArg):
    def __init__(self, name, index, type, default_value, label=None):
        super().__init__(name, index, type, default_value)
        self.docs = None
        self.label = label or name

    def __str__(self):
        return f"{self.label}[{self.index}]: {self.type} = {self.default_value}"

    def to_schema_info(self):
        type, default = type_to_schema_type(self.type, self.default_value)
        return {
            "type": type,
            "default": default,
            "description": self.docs,
        }

    def to_ui_info(self, toolbox):
        tag_value = toolbox.enum_class_to_wrapper.get(self.type)
        if tag_value:
            return { "prefix": self.label, "dtypeName": tag_value.id }
        else:
            return { "prefix": self.label }

    def apply_overrides(self, d):
        name = d.get("name")
        docs = d.get("docs")

        if name:
            self.label = name

        if docs:
            self.docs = docs


def type_to_schema_type(t, default):
    if t is float:
        return "number", default if default is not None else 0.0
    elif t is str:
        return "string", default if default is not None else ""
    elif t is bool:
        return "boolean", default if default is not None else False
    elif t is int:
        return "integer", default if default is not None else 0
    else:
        if not (isinstance(t, type) and issubclass(t, (Enum, DynEnum))):
            print("unknown type for schema", t, default)
        return "string", ""


class FnInfo:
    def __init__(self, fn, name, docs, args, return_type=None):
        self.fn = fn
        self.name = name
        self.docs = docs
        self.args = args
        self.return_type = return_type

    def __str__(self):
        return f"def {self.name}({', '.join([str(arg) for arg in self.args])}) -> {self.return_type}:\n\t{self.docs}"

    async def call_with_args(self, args_provider, info):
        args0 = [args_provider.provide_arg(arg, info) for arg in self.args]
        # await all items in arg that needs awaiting
        args = [await arg if inspect.isawaitable(arg) else arg for arg in args0]
        return self.fn(*args)

    @classmethod
    def make_arg(cls, name, i, arg_type, arg_default):
        return FnArg(name, i, arg_type, arg_default)

    @classmethod
    def from_function(cls, fn):
        fn_name = fn.__name__
        docs = fn.__doc__
        default_count = len(fn.__defaults__ or [])
        arg_names = fn.__code__.co_varnames[: fn.__code__.co_argcount]
        arg_types = {}
        arg_defaults = {}
        return_type = None

        args = []

        if fn.__defaults__:
            default_params = arg_names[-default_count:]
            for i, param in enumerate(default_params):
                arg_defaults[param] = fn.__defaults__[i]

        for param, annotation in fn.__annotations__.items():
            if param == "return":
                return_type = annotation
            else:
                arg_types[param] = annotation

        for i, name in enumerate(arg_names):
            arg_type = arg_types.get(name)
            arg_default = arg_defaults.get(name)
            args.append(cls.make_arg(name, i, arg_type, arg_default))

        return cls(fn, fn_name, docs, args, return_type)


class Tool(FnInfo):
    def __init__(self, fn, name, docs, args, return_type=None, examples=None):
        super().__init__(fn, name, docs, args, return_type)
        self.id = name
        self.examples = examples if examples is not None else []
        self.ui_prefix = name

    @classmethod
    def make_arg(cls, name, i, arg_type, arg_default):
        return ToolArg(name, i, arg_type, arg_default)


    def to_info(self, toolbox):
        args = [arg for arg in self.args if arg.type is not toolbox.State]
        return {
            "title": self.name,
            "description": self.docs,
            "schema": {"fields": {arg.name: arg.to_schema_info() for arg in args}},
            "ui": {
                "prefix": self.ui_prefix,
                "args": {arg.name: arg.to_ui_info(toolbox) for arg in args},
            },
            "examples": self.examples,
        }

    def apply_overrides(self, d):
        id = d.get("id")
        name = d.get("name")
        docs = d.get("docs")
        args = d.get("args")
        examples = d.get("examples")
        ui_prefix = d.get("ui_prefix")

        if id:
            self.id = id

        if name:
            self.name = name

        if docs:
            self.docs = docs

        if args:
            args_by_name = {arg.name: arg for arg in self.args}
            for name, info in args.items():
                arg = args_by_name.get(name)
                if arg:
                    if isinstance(info, dict):
                        arg.apply_overrides(info)
                    elif isinstance(info, str):
                        arg.apply_overrides({"name": info})
                    else:
                        print("bad tool argument info format for argument", name, info)

                else:
                    print("tool argument info for inexistent argument name", name)

        if examples:
            self.examples = examples

        if ui_prefix:
            self.ui_prefix = ui_prefix
        elif self.ui_prefix == self.id:
            self.ui_prefix = self.name


def basic_match_raw_tag_value(word, possibilities):
    s = SequenceMatcher()
    s.set_seq2(word)

    best_key_index = -1
    best_key_ratio = -1
    best_key_pair = None

    best_value_index = -1
    best_value_ratio = -1
    best_value_pair = None

    for i, pair in enumerate(possibilities):
        key, value = pair

        s.set_seq1(key)
        key_ratio = s.ratio()
        if key_ratio > best_key_ratio:
            best_key_ratio = key_ratio
            best_key_index = i
            best_key_pair = pair

        s.set_seq1(value)
        value_ratio = s.ratio()
        if value_ratio > best_value_ratio:
            best_value_ratio = value_ratio
            best_value_index = i
            best_value_pair = pair

    if best_key_ratio > best_value_ratio:
        return best_key_index, best_key_pair
    else:
        return best_value_index, best_value_pair


def to_seq_of_pairs(seq):
    for k, v in seq:
        yield (k, v)


def to_list_of_pairs(seq):
    return list(to_seq_of_pairs(seq))


class DynEnum:
    def __init__(self, key, label):
        self.key = key
        self.label = label

    def __eq__(self, other):
        return self.key == other.key and self.label == other.label

    @staticmethod
    def load():
        return []


class TagValue:
    def __init__(self, id, name, docs, icon=None):
        self.id = id
        self.name = name
        self.docs = docs
        self.icon = None
        self.closest_matcher = basic_match_raw_tag_value

    def __str__(self):
        return f"TagValue({self.id}, {self.name}, {self.docs})"

    def to_info(self, _toolbox):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.docs,
            "icon": self.icon,
        }

    async def from_raw_arg_value(self, v):
        return await self.closest_match(str(v))

    async def match_handler(self, info):
        value = info.get("value", "")
        return await self.closest_match(value)

    async def load_handler(self, _info):
        entries = to_list_of_pairs(await self.get_variants())
        return dict(entries=entries)

    async def closest_match(self, word):
        variant_pairs = to_seq_of_pairs(await self.get_variants())
        _, pair = self.closest_matcher(word, variant_pairs)
        return pair

    async def get_variants(self):
        return []

    def apply_overrides(self, d):
        id = d.get("id")
        name = d.get("name")
        docs = d.get("docs")
        icon = d.get("icon")
        matcher = d.get("matcher")

        if id:
            self.id = id

        if name:
            self.name = name

        if docs:
            self.docs = docs

        if icon:
            self.icon = icon

        if matcher:
            self.closest_matcher = matcher

    @classmethod
    def from_enum_class(cls, Class, overrides):
        id = Class.__name__
        name = id
        docs = Class.__doc__
        variants = []
        for variant in Class:
            variants.append(Variant.from_enum_variant(variant))

        e = FixedTagValue(id, name, docs, variants, Class)
        e.apply_overrides(overrides)

        return e

    @classmethod
    def from_dyn_enum_class(cls, Class, overrides, fn_arg_provider):
        id = Class.__name__
        name = id
        docs = Class.__doc__

        load_fn_info = FnInfo.from_function(Class.load)

        e = DynTagValue(
            id,
            name,
            docs,
            Class,
            lambda: load_fn_info.call_with_args(fn_arg_provider, {}),
        )
        e.apply_overrides(overrides)

        return e


class FixedTagValue(TagValue):
    def __init__(self, id, name, docs, variants, EnumClass):
        super().__init__(id, name, docs)
        self.EnumClass = EnumClass
        self.variants = variants

    async def from_raw_arg_value(self, v):
        v = await self.closest_match(str(v))
        if v:
            key, value = v
            return self.EnumClass.__members__.get(key)

        return None

    @property
    def match_handler_id(self):
        return f"enum::{self.id}::match"

    def to_info(self, _toolbox):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.docs,
            "icon": self.icon,
            "entries": [(v.name, v.value) for v in self.variants],
            "matchHandlerId": self.match_handler_id,
        }

    def get_handlers(self):
        return [
            (self.match_handler_id, self.match_handler),
        ]

    async def get_variants(self):
        return self.variants


class DynTagValue(TagValue):
    def __init__(self, id, name, docs, EnumClass, load_fn, cache=True):
        super().__init__(id, name, docs)
        self.EnumClass = EnumClass
        self.load_fn = load_fn
        self.cache = cache
        self._cached_variants = None

    async def from_raw_arg_value(self, v):
        v = await self.closest_match(str(v))
        if v:
            key, value = v
            return self.EnumClass(key, value)

        return None

    @property
    def match_handler_id(self):
        return f"enum::{self.id}::match"

    @property
    def load_handler_id(self):
        return f"enum::{self.id}::load_entries"

    async def get_variants(self):
        if self._cached_variants:
            return self._cached_variants

        variants = await maybe_await(await self.load_fn())

        if self.cache:
            self._cached_variants = variants

        return variants

    def get_handlers(self):
        return [
            (self.match_handler_id, self.match_handler),
            (self.load_handler_id, self.load_handler),
        ]

    def to_info(self, _toolbox):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.docs,
            "icon": self.icon,
            "matchHandlerId": self.match_handler_id,
            "loadEntriesHandlerId": self.load_handler_id,
        }


class Variant:
    def __init__(self, name, value, docs=None):
        self.name = name
        self.value = value
        self.docs = docs

    def __str__(self):
        return f"Variant({self.name}, {self.value}, {self.docs})"

    def __iter__(self):
        yield self.name
        yield self.value

    @classmethod
    def from_enum_variant(cls, variant):
        name = variant.name
        value = variant.value
        docs = None

        return cls(name, value, docs)


class EmptyState:
    "State class used when no state is provided"

    def setup(self):
        "method called to setup the state before start serving"
        pass


class Toolbox:
    def __init__(self, id, name, docs, state=None):
        self.id = id
        self.name = name
        self.docs = docs

        self.tools = []
        self.enums = []
        self.enum_class_to_wrapper = {}

        self.handlers = {}

        self.state = state if state is not None else EmptyState()
        self.State = self.state.__class__

    def __str__(self):
        if self.tools:
            tools = "\n\n" + "\n\n".join([str(tool) for tool in self.tools])
        else:
            tools = ""

        return f"Toolbox({self.id}, {self.name}, {self.docs}){tools}"

    def provide_arg(self, arg, info):
        if arg.type is self.State:
            return self.state
        else:
            v = info.get(arg.name)
            if isinstance(arg.type, type) and issubclass(arg.type, (Enum, DynEnum)):
                wrapper = self.enum_class_to_wrapper.get(arg.type)
                if wrapper:
                    # NOTE: This call is async and returns an awaitable
                    return wrapper.from_raw_arg_value(v)
                else:
                    print("enum class argument type not annotated?", arg)
                    return arg.default_value
            else:
                return arg.cast_json_value_or_default(v)

    async def setup(self):
        if hasattr(self.state, "setup"):
            if inspect.iscoroutinefunction(self.state.setup):
                await self.state.setup()
            else:
                self.state.setup()

    def add_enum(self, v):
        for id, handler in v.get_handlers():
            if id in self.handlers:
                print(f"WARN: duplicated handler for id '{id}, new for enum {v}")

            self.handlers[id] = handler

        self.enum_class_to_wrapper[v.EnumClass] = v
        self.enums.append(v)

    def enum(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            arg = args[0]
            if isinstance(arg, type) and issubclass(arg, Enum):
                self.add_enum(TagValue.from_enum_class(arg, {}))
            elif isinstance(arg, type) and issubclass(arg, DynEnum):
                self.add_enum(TagValue.from_dyn_enum_class(arg, {}, self))
            else:
                raise TypeError(
                    "The decorated class must inherit from enum.Enum or DynEnum"
                )

            return arg
        else:

            def wrapper(arg):
                if isinstance(arg, type) and issubclass(arg, Enum):
                    self.add_enum(TagValue.from_enum_class(arg, kwargs))
                elif isinstance(arg, type) and issubclass(arg, DynEnum):
                    self.add_enum(TagValue.from_dyn_enum_class(arg, kwargs, self))
                else:
                    raise TypeError(
                        "The decorated class must inherit from enum.Enum or DynEnum"
                    )

                return arg

            return wrapper

    def add_tool(self, tool):
        if tool.id in self.handlers:
            print(f"WARN: duplicated handler for id '{id}, new for tool {tool}")

        arg_provider = self
        self.handlers[tool.id] = lambda info: tool.call_with_args(arg_provider, info)

        self.tools.append(tool)

    def tool(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            func = args[0]
            t = Tool.from_function(func)
            self.add_tool(t)
            return func
        else:

            def wrapper(func):
                t = Tool.from_function(func)
                t.apply_overrides(kwargs)
                self.add_tool(t)
                return func

            return wrapper

    def build_tool_info(self):
        return {
            "ns": self.id,
            "title": self.name,
            "description": self.docs,
            "tools": {tool.id: tool.to_info(self) for tool in self.tools},
            "tagValues": {
                tag_value.id: tag_value.to_info(self) for tag_value in self.enums
            },
        }

    def handle_request(self, body):
        if not isinstance(body, dict):
            return res_error("BadRequestBody", "Bad Request Body")

        action = body.get("action", None)
        if action == "info":
            return self.handle_action_info()
        elif action == "request":
            op_name = body.get("opName", None)
            req_info = body.get("info", {})
            return self.handle_action_request(op_name, req_info)
        else:
            return res_error("UnknownAction", "Unknown Action", {"action": action})

    def handle_action_info(self):
        return self.build_tool_info()

    def handle_action_request(self, op_name, req_info):
        handler = self.handlers.get(op_name)
        if handler:
            return handler(req_info)
        else:
            return res_error("ToolNotFound", "Tool Not Found", {"opName": op_name})

    def to_fastapi_app(self):
        from fastapi import FastAPI, Request
        from fastapi.encoders import jsonable_encoder
        from fastapi.responses import JSONResponse

        app = FastAPI()
        tb = self

        @app.post("/")
        async def root_gd_handler(request: Request):
            body = await request.json()
            res = await maybe_await(tb.handle_request(body))
            try:
                return JSONResponse(content=jsonable_encoder(res), status_code=200)
            except Exception as err:
                print("Error encoding response", err, res)
                err_res = res_error("InternalError", "Internal Error", {})
                return JSONResponse(content=jsonable_encoder(err_res), status_code=500)

        return app

    def serve(self, host="127.0.0.1", port=8086):
        loop = asyncio.get_event_loop()
        if loop.is_closed():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        loop.run_until_complete(self.setup())
        serve_uvicorn(self.to_fastapi_app(), host, port)


def res_error(code, reason, info=None):
    return {"ok": False, "code": code, "reason": reason, "info": info}


def serve_uvicorn(app, host="127.0.0.1", port=8086):
    import uvicorn

    return uvicorn.run(app, host=host, port=port)


async def maybe_await(v):
    if inspect.isawaitable(v):
        return await v
    else:
        return v
